/*! \file TransitionMatrix.h
 *
 * \brief Representation of the Transition Matrix.
 */

#ifndef INCLUDE_TRANSITIONMATRIX_H_
#define INCLUDE_TRANSITIONMATRIX_H_

/**
 * \brief Exception for unrecognized file formats.
 */
class UnrecognizedFormatException: public std::exception {
protected:
  //! Description of the problem.
  std::string message;
public:
  /**
   * \brief Exception instance constructor.
   *
   * \param problem: `string` Description of the problem that occurred.
   */
  UnrecognizedFormatException(std::string problem) { message = problem; }
  /**
   * \brief Exception message.
   */
  virtual const char* what() const throw() {
    return message.c_str();
  }
};

/*! \brief Class to represent the Transition Matrix.
 */
class TransitionMatrix {
 protected:
  //! Matrix type identifier.
  int is;
  //! Maximum field expansion order.
  int l_max;
  //! Wave number in scale units.
  double vk;
  //! External medium refractive index.
  double exri;
  //! Vectorized matrix elements.
  std::complex<double> *elements;
  //! Sphere radius.
  double sphere_radius;
  //! Matrix shape
  int *shape;

  /*! \brief Build transition matrix from a HDF5 binary input file.
   *
   * This function takes care of the specific task of building a transition
   * matrix memory data structure from a HDF5 binary input file.
   *
   * \param file_name: `string` Name of the binary configuration data file.
   * \return config: `TransitionMatrix *` Pointer to object containing the
   * transition matrix data.
   */
  static TransitionMatrix *from_hdf5(std::string file_name);
  
  /*! \brief Build transition matrix from a legacy binary input file.
   *
   * \param file_name: `string` Name of the binary configuration data file.
   * \return config: `TransitionMatrix *` Pointer to object containing the
   * transition matrix data.
   */
  static TransitionMatrix *from_legacy(std::string file_name);

  /*! \brief Write the Transition Matrix to HDF5 binary output.
   *
   * This function takes care of the specific task of building a transition
   * matrix memory data structure from a binary input file formatted according
   * to the structure used by the original FORTRAN code.
   *
   * \param file_name: `string` Name of the binary configuration data file.
   */
  void write_hdf5(std::string file_name);
  
  /*! \brief Write the Transition Matrix to legacy binary output.
   *
   * \param file_name: `string` Name of the binary configuration data file.
   */
  void write_legacy(std::string file_name);
 public:
  /*! \brief Default Transition Matrix instance constructor.
   *
   * \param _is: `int` Matrix type identifier
   * \param _lm: `int` Maximum field expansion order.
   * \param _vk: `double` Wave number in scale units.
   * \param _exri: `double` External medium refractive index.
   * \param _elements: `complex<double> *` Vectorized elements of the matrix.
   * \param _radius: `double` Radius for the single sphere case (defaults to 0.0).
   */
  TransitionMatrix(
		   int _is, int _lm, double _vk, double _exri, std::complex<double> *_elements,
		   double _radius=0.0
  );

  /*! \brief Transition Matrix instance constructor for single sphere.
   *
   * This constructor allocates the memory structure needed to represent the transition
   * matrix for the case of a single sphere.
   *
   * \param _lm: `int` Maximum field expansion order.
   * \param _vk: `double` Wave number in scale units.
   * \param _exri: `double` External medium refractive index.
   * \param _rmi: Matrix of complex.
   * \param _rei: Matrix of complex.
   * \param _sphere_radius: `double` Radius of the sphere.
   */
  TransitionMatrix(
		   int _lm, double _vk, double _exri, std::complex<double> **_rmi,
		   std::complex<double> **_rei, double _sphere_radius
  );

  /*! \brief Transition Matrix instance constructor for a cluster of spheres.
   *
   * This constructor allocates the memory structure needed to represent the transition
   * matrix for the case of a cluster of spheres.
   *
   * \param _nlemt: `int` Size of the matrix (2 * LE * (LE + 2)).
   * \param _lm: `int` Maximum field expansion order.
   * \param _vk: `double` Wave number in scale units.
   * \param _exri: `double` External medium refractive index.
   * \param _am0m: Matrix of complex.
   */
  TransitionMatrix(int _nlemt, int _lm, double _vk, double _exri, std::complex<double> **_am0m);

  /*! \brief Transition Matrix instance destroyer.
   */
  ~TransitionMatrix();
  
  /*! \brief Build transition matrix from binary input file.
   *
   * In some cases, it is necessary to perform calculations starting from a pre-computed
   * transition matrix. If this matrix is not available in memory (e.g. because it was
   * calculated by a different process), but it was saved in a binary file, this function
   * can be used to load it back in memory. The function operates in two modes: "LEGACY",
   * which reads the matrix data from a proprietary binary file, having the same structure
   * as the one used by the original FORTRAN code, and "HDF5", which, instead, reads the
   * data from an input file complying with the HDF5 format.
   *
   * \param file_name: `string` Name of the binary configuration data file.
   * \param mode: `string` Binary encoding. Can be one of ["LEGACY", "HDF5"]. Optional
   * (default is "LEGACY").
   * \return config: `TransitionMatrix *` Pointer to object containing the transition
   * matrix data.
   */
  static TransitionMatrix* from_binary(std::string file_name, std::string mode="LEGACY");

  /*! \brief Write the Transition Matrix to a binary file.
   *
   * This function writes a hard-copy of the transition matrix to an output file, making
   * it available for subsequent processes to reload. The function operates in two modes:
   * "LEGACY", which writes a proprietary binary file, using the same structure of the
   * original FORTRAN code, and "HDF5", which, instead, writes the output to a file using
   * the HDF5 format, thus leaving it available for inspection with external tools.
   *
   * \param file_name: `string` Name of the file to be written.
   * \param mode: `string` Binary encoding. Can be one of ["LEGACY", "HDF5"] . Optional
   * (default is "LEGACY").
   */
  void write_binary(std::string file_name, std::string mode="LEGACY");
  
  /*! \brief Test whether two instances of TransitionMatrix are equal.
   *
   * Transition matrices can be the result of a calculation or of a file input operation,
   * reading from a previously computed object. The `==` operator tests for the equivalence
   * of two transition matrices, returning `true` if they are equivalnet, `false` otherwise.
   *
   * \param other: `TransitionMatrix &` Reference to the instance to be compared with.
   * \return result: `bool` True, if the two instances are equal, false otherwise.
   */
  bool operator ==(TransitionMatrix &other);
};

#endif
